@@ -24,7 +24,7 @@ from torch import nn, Tensor, tensor, is_tensor
 import torch.nn.functional as F
 from torch.nn import Module, ModuleList, Linear
 from torch.nn.utils.rnn import pad_sequence
-from torch.utils._pytree import tree_map
+from torch.utils._pytree import tree_map, tree_flatten, tree_unflatten
 
 from torchdiffeq import odeint
 
@@ -32,6 +32,8 @@ import einx
 from einops import rearrange, repeat, reduce, einsum, pack
 from einops.layers.torch import Rearrange
 
+from loguru import logger
+
 from ema_pytorch import EMA
 
 from rotary_embedding_torch import RotaryEmbedding, apply_rotary_emb
@@ -72,7 +74,7 @@ except ImportError:
 
 # constants
 
-ModalitySample = list[Int['_'] | Float['...'] | tuple[int, Float['_ _']]]
+ModalitySample = list[Int[''] | Int['_'] | Float['...'] | tuple[int, Float['...']]]
 
 ModalityTokenTransform = str | Callable | None
 
@@ -83,6 +85,7 @@ class LossBreakdown(NamedTuple):
     text: Float['']
     flow: list[Float['']]
     velocity: list[Float['']] | None
+    direction: list[Float['']] | None
 
 # helper functions
 
@@ -98,9 +101,6 @@ def identity(t):
 def first(it):
     return it[0]
 
-def prepend(arr, el):
-    arr.insert(0, el)
-
 def join(arr, delimiter = ''):
     return delimiter.join(arr)
 
@@ -143,11 +143,11 @@ def default_modality_length_to_time_fn(modality_length: Int['b m']) -> Float['b
     rand_num_modalities = torch.floor(torch.rand_like(num_modalities) * num_modalities)
     seq = torch.arange(total_modalities, device = device)
 
-    prev_decoded_modality = einx.less_equal('n, b -> b n', seq, rand_num_modalities)
+    prev_decoded_modality = einx.less('m, b -> b m', seq, rand_num_modalities)
     curr_modality_rand_time = torch.rand_like(num_modalities)
 
     # in paper, they fix previous decoded modalities to 500 / 1000 steps for discrete ddpm, here using flow matching with times 0 - 1 so corresponds to 0.5
-    return torch.where(prev_decoded_modality, 0.5, curr_modality_rand_time)
+    return einx.where('b m, , b -> b m', prev_decoded_modality, 0.5, curr_modality_rand_time)
 
 # pretty print
 
@@ -209,6 +209,20 @@ def l2norm(t):
 def softclamp(t, value = 50.):
     return (t / value).tanh() * value
 
+def max_neg_value(t):
+    return -torch.finfo(t.dtype).max
+
+def log(t, eps = 1e-20):
+    return torch.log(t.clamp(min = eps))
+
+def gumbel_noise(t):
+    noise = torch.rand_like(t)
+    return -log(-log(noise))
+
+def gumbel_sample(t, temperature = 1., dim = -1, keepdim = True):
+    noise = gumbel_noise(t) * int(temperature > 0)
+    return (t / temperature + noise).argmax(dim = dim, keepdim = keepdim)
+
 # flex attention mask construction
 # https://pytorch.org/blog/flexattention/
 
@@ -245,6 +259,11 @@ def softcap_score_mod(softcap):
         return score
     return inner
 
+# losses
+
+def calc_direction_loss(pred, target):
+    return 0.5 * (1. - einsum(l2norm(pred), l2norm(target), '... d, ... d -> ...'))
+
 # converting a raw list of modality offsets and lengths to tensor
 
 @typecheck
@@ -507,22 +526,17 @@ class AdaptiveWrapper(Module):
         x: Float['b n {self.dim}'],
         **kwargs
     ):
-
         x = self.layernorm(x)
 
         x = x * (self.layernorm_gamma + 1.)
 
         out = self.fn(x, **kwargs)
 
-        multiple_returns = isinstance(out, tuple)
-
-        if multiple_returns:
-            out, *rest = out
+        (out, *rest), tree_spec = tree_flatten(out)
 
         out = out * (self.layerscale + 1.)
 
-        if multiple_returns:
-            out = (out, *rest)
+        out = tree_unflatten((out, *rest), tree_spec)
 
         return out
 
@@ -530,7 +544,7 @@ class AdaptiveWrapper(Module):
     def forward_modality(
         self,
         x: Float['b n {self.dim}'],
-        cond: Float['b {self.dim_cond}'] | Float['b n {self.dim_cond}'],
+        cond: Float['... {self.dim_cond}'],
         **kwargs
     ):
         x = self.layernorm(x)
@@ -541,23 +555,17 @@ class AdaptiveWrapper(Module):
 
         # attention or feedforwards
 
-        out = self.fn(x, **kwargs)
+        out = self.fn(modality_tokens, **kwargs)
 
-        multiple_returns = isinstance(out, tuple)
-
-        if multiple_returns:
-            out, *rest = out
+        (out, *rest), tree_spec = tree_flatten(out)
 
         # take care of conditioning output separately for text vs modality
 
         modalities_out = out * self.to_ada_ln_zero(cond).sigmoid()
 
-        # take care of function returning cache
+        # take care of function returning cache or value residual
 
-        if not multiple_returns:
-            return modalities_out
-
-        return (modalities_out, *rest)
+        modalities_out = tree_unflatten((modalities_out, *rest), tree_spec)
 
         return modalities_out
 
@@ -565,7 +573,7 @@ class AdaptiveWrapper(Module):
     def forward(
         self,
         x: Float['b n {self.dim}'],
-        cond: Float['b {self.dim_cond}'] | Float['b n {self.dim_cond}'] | None = None,
+        cond: Float['... {self.dim_cond}'] | None = None,
         is_any_modality: bool | Bool['b n'] | None = None,
         modality_only = False,
         **kwargs
@@ -602,10 +610,7 @@ class AdaptiveWrapper(Module):
 
         out = self.fn(x, **kwargs)
 
-        multiple_returns = isinstance(out, tuple)
-
-        if multiple_returns:
-            out, *rest = out
+        (out, *rest), tree_spec = tree_flatten(out)
 
         # take care of conditioning output separately for text vs modality
 
@@ -615,12 +620,11 @@ class AdaptiveWrapper(Module):
 
         conditioned_out = torch.where(is_any_modality, modalities_out, text_out)
 
-        # take care of function returning cache
+        # take care of function returning cache or value residual
 
-        if not multiple_returns:
-            return conditioned_out
+        conditioned_out = tree_unflatten((conditioned_out, *rest), tree_spec)
 
-        return (conditioned_out, *rest)
+        return conditioned_out
 
 # attention
 
@@ -645,7 +649,6 @@ def FeedForward(
 ):
     dim_inner = int(dim * expansion_factor * 2 / 3)
     return nn.Sequential(
-        RMSNorm(dim),
         Linear(dim, dim_inner * 2),
         GEGLU(),
         nn.Dropout(dropout),
@@ -670,8 +673,6 @@ class Attention(Module):
         assert not (use_flex_attn and not exists(flex_attention)), 'flex attention is only available on torch 2.5.0 (nightly) onwards'
         self.use_flex_attn = use_flex_attn
 
-        self.norm = RMSNorm(dim)
-
         self.to_qkv = nn.Sequential(
             Linear(dim, dim_inner * 3, bias = False),
             Rearrange('b n (qkv h d) -> qkv b h n d', qkv = 3, h = heads)
@@ -699,7 +700,9 @@ class Attention(Module):
         cache: Tensor | None = None,
         causal = False,
         block_mask = None, # only passed in for flex attention
-        return_kv_cache = False
+        return_kv_cache = False,
+        return_values = False,
+        value_residual: Tensor | None = None
     ):
         device, input_is_cuda, is_decoding_with_cache = x.device, x.is_cuda, exists(cache)
 
@@ -714,12 +717,17 @@ class Attention(Module):
         assert not (exists(block_mask) and exists(attn_mask))
         assert not (not self.use_flex_attn and exists(block_mask)), 'you cannot pass in the `block_mask` if `use_flex_attn` was not set to be `True`'
 
-        # pre rmsnorm and project to queries, keys, values
-
-        x = self.norm(x)
+        # project to queries, keys, values
 
         q, k, v = self.to_qkv(x)
 
+        # value residual
+
+        orig_v = v
+
+        if exists(value_residual):
+            v = 0.5 * (v + value_residual)
+
         # handle cache being passed in
 
         if exists(cache):
@@ -755,7 +763,7 @@ class Attention(Module):
 
             sim = softclamp(sim, self.softcap_value)
 
-            mask_value = -torch.finfo(sim.dtype).max
+            mask_value = max_neg_value(sim)
 
             if causal:
                 i, j = sim.shape[-2:]
@@ -780,6 +788,9 @@ class Attention(Module):
 
         out = self.to_out(out)
 
+        if return_values:
+            out = (out, orig_v)
+
         if not return_kv_cache:
             return out
 
@@ -919,6 +930,8 @@ class Transformer(Module):
         # transformer layers as usual, using mask from above
 
         skips = []
+        value_residual = None
+
         new_cache = []
 
         depth = len(self.layers)
@@ -943,15 +956,19 @@ class Transformer(Module):
 
             # attention and feedforward
 
-            attn_out, kv_cache = attn(
+            (attn_out, attn_values), kv_cache = attn(
                 x,
                 rotary_emb = rotary_emb,
                 cache = next(iter_cache, None),
                 return_kv_cache = True,
+                return_values = True,
+                value_residual = value_residual,
                 **attn_mask_kwargs,
                 **adaptive_kwargs
             )
 
+            value_residual = default(value_residual, attn_values)
+
             new_cache.append(kv_cache)
 
             x = attn_out + x
@@ -987,6 +1004,8 @@ class Transfusion(Module):
         to_modality_shape_fn: Callable | tuple[Callable, ...] = default_to_modality_shape_fn,
         ignore_index = -1,
         flow_loss_weight = 1.,
+        add_flow_direction_loss = False, # figure 7 https://arxiv.org/abs/2410.10356
+        direction_loss_weight = 1.,
         velocity_consistency_loss_weight = 1.,
         odeint_kwargs: dict = dict(
             atol = 1e-5,
@@ -1125,6 +1144,9 @@ class Transfusion(Module):
 
         self.to_text_logits = Linear(dim, effective_num_text_tokens, bias = False)
 
+        text_only_mask = torch.arange(effective_num_text_tokens) < num_text_tokens
+        self.register_buffer('text_only_logits_mask', text_only_mask, persistent = False)
+
         self.model_to_latent_preds = ModuleList([Linear(dim, dim_latent, bias = False) for dim_latent in self.dim_latents])
 
         # loss related
@@ -1136,6 +1158,11 @@ class Transfusion(Module):
 
         self.velocity_consistency_loss_weight = velocity_consistency_loss_weight
 
+        # flow direction loss related
+
+        self.add_flow_direction_loss = add_flow_direction_loss
+        self.direction_loss_weight = direction_loss_weight
+
         # flow sampling related
 
         self.odeint_fn = partial(odeint, **odeint_kwargs)
@@ -1195,7 +1222,6 @@ class Transfusion(Module):
 
             if exists(fixed_modality_shape):
                 modality_shape = fixed_modality_shape
-                return
 
             # get the tokens after the modality meta id
 
@@ -1212,7 +1238,7 @@ class Transfusion(Module):
 
                 if not meta_str.isdigit() or int(meta_str) <= 0:
 
-                    assert exists(default_shape), f'invalid modality meta information detected, please set `modality_default_shape` in order to properly fallback'
+                    assert exists(default_shape), 'invalid modality meta information detected, please set `modality_default_shape` in order to properly fallback'
                     modality_shape = default_shape
                 else:
                     modality_shape = meta_str_to_modality_shape(meta_str)
@@ -1275,6 +1301,7 @@ class Transfusion(Module):
                     sampled_token_id = sampled.item()
 
                     if sampled_token_id == self.eos_id:
+                        logger.info(f'detecting an end of string token [{self.eos_id}], terminating sampling early')
                         break
 
                     maybe_transition_to_modality_decoding(seq)
@@ -1368,6 +1395,7 @@ class Transfusion(Module):
                 continue
 
             modality_id, modality = sample
+
             maybe_modality_decoder = self.modality_decoder[modality_id]
 
             if self.channel_first_latent:
@@ -1397,7 +1425,8 @@ class Transfusion(Module):
         device = self.device
         text = text.to(device)
 
-        text, labels = text[:, :-1], text[:, 1:]
+        if return_loss:
+            text, labels = text[:, :-1], text[:, 1:]
 
         # embed text
 
@@ -1431,6 +1460,8 @@ class Transfusion(Module):
 
             return logits, kv_cache
 
+        logits = logits.masked_fill(~self.text_only_logits_mask, max_neg_value(logits))
+
         loss = F.cross_entropy(
             rearrange(logits, 'b n l -> b l n'),
             labels,
@@ -1439,15 +1470,45 @@ class Transfusion(Module):
 
         return loss
 
+    @torch.no_grad()
+    @eval_decorator
+    @typecheck
+    def generate_text_only(
+        self,
+        prompt: Int['b n'],
+        seq_len: int,
+        temperature = 1.5,
+        min_p = 0.1,
+    ) -> Int['b no']:
+
+        prompt_seq_len, out = prompt.shape[-1], prompt.clone()
+        sample_num_times = max(0, seq_len - prompt_seq_len)
+
+        for _ in tqdm(range(sample_num_times)):
+            logits = self.forward_text(out, return_loss = False)
+            logits = logits[:, -1]
+
+            logits = min_p_filter(logits, min_p = min_p)
+
+            logits.masked_fill_(~self.text_only_logits_mask, max_neg_value(logits))
+
+            sample = gumbel_sample(logits, temperature = temperature, dim = -1)
+
+            out = torch.cat((out, sample), dim = -1)
+
+        return out[..., prompt_seq_len:]
+
     @typecheck
     def forward_modality(
         self,
         modalities: Float['b ...'],
+        times: Float['b'] | None = None,
         modality_type: int | None = None,
+        encode_modality: bool = True,
         return_loss = True
     ) -> Float['']:
 
-        shape = modalities.shape
+        modalities = modalities.to(self.device)
 
         if self.num_modalities > 1:
             assert exists(modality_type), '`modality_type` must be explicitly passed in on forward when training on greater than 1 modality'
@@ -1455,6 +1516,7 @@ class Transfusion(Module):
         modality_type = default(modality_type, 0)
 
         transform = self.modality_token_transform[modality_type]
+        maybe_modality_encode = self.modality_encoder[modality_type]
         latent_to_model_fn = self.latent_to_model_projs[modality_type]
         model_to_flow_pred_fn = self.model_to_latent_preds[modality_type]
 
@@ -1464,6 +1526,17 @@ class Transfusion(Module):
         maybe_pos_emb_mlp = self.pos_emb_mlp[modality_type]
         modality_num_dim = self.modality_num_dim[modality_type]
 
+        # maybe modality encode
+
+        if encode_modality and exists(maybe_modality_encode):
+            with torch.no_grad():
+                maybe_modality_encode.eval()
+                modalities = maybe_modality_encode(modalities).detach()
+
+        shape = modalities.shape
+
+        # axial positions
+
         if add_pos_emb:
             assert exists(modality_num_dim), f'modality_num_dim must be set for modality {modality_type} if further injecting axial positional embedding'
 
@@ -1495,7 +1568,9 @@ class Transfusion(Module):
 
         # times
 
-        times = torch.rand((batch,), device = device)
+        if not exists(times):
+            times = torch.rand((batch,), device = device)
+
         padded_times = rearrange(times, 'b -> b 1 1')
 
         noise = torch.randn_like(tokens)
@@ -1510,7 +1585,6 @@ class Transfusion(Module):
 
         if add_pos_emb:
             axial_pos_emb = maybe_pos_emb_mlp(tensor(axial_dims), flatten_dims = True)
-
             noised_tokens = noised_tokens + axial_pos_emb
 
         # attention
@@ -1531,6 +1605,76 @@ class Transfusion(Module):
 
         return F.mse_loss(pred_flow, flow)
 
+    @torch.no_grad()
+    @eval_decorator
+    @typecheck
+    def generate_modality_only(
+        self,
+        batch_size: int = 1,
+        modality_type: int | None = None,
+        fixed_modality_shape: tuple[int, ...] | None = None,
+        modality_steps = 16,
+        return_unprocessed_modalities = False
+    ) -> Tensor:
+
+        device = self.device
+
+        if self.num_modalities > 1:
+            assert exists(modality_type), '`modality_type` must be explicitly passed in on forward when training on greater than 1 modality'
+
+        modality_type = default(modality_type, 0)
+
+        latent_dim = self.dim_latents[modality_type]
+        to_flow_pred = self.model_to_latent_preds[modality_type]
+        maybe_modality_decoder = self.modality_decoder[modality_type]
+        maybe_modality_decoder = self.modality_decoder[modality_type]
+        default_modality_shape = self.modality_default_shape[modality_type]
+
+        modality_shape = default(fixed_modality_shape, default_modality_shape)
+
+        assert exists(modality_shape)
+
+        axial_dims = modality_shape
+        modality_length = math.prod(modality_shape)
+
+        noise = torch.randn((batch_size, modality_length, latent_dim), device = device)
+
+        def ode_step_fn(step_times, denoised):
+
+            step_times = repeat(step_times, ' -> b', b = batch_size)
+
+            denoised = denoised.reshape(batch_size, *modality_shape, latent_dim)
+
+            flow = self.forward_modality(
+                denoised,
+                times = step_times,
+                encode_modality = False,
+                return_loss = False
+            )
+
+            return flow
+
+        times = torch.linspace(0., 1., modality_steps, device = device)
+        trajectory = self.odeint_fn(ode_step_fn, noise, times)
+
+        # add the sampled modality tokens
+
+        sampled_modality = trajectory[-1]
+
+        # reshape
+
+        sampled_modality = sampled_modality.reshape(batch_size, *modality_shape, latent_dim)
+
+        # decode
+
+        if self.channel_first_latent:
+            sampled_modality = rearrange(sampled_modality, 'b ... d -> b d ...')
+
+        if exists(maybe_modality_decoder):
+            sampled_modality = maybe_modality_decoder(sampled_modality)
+
+        return sampled_modality
+
     @typecheck
     def forward(
         self,
@@ -1605,9 +1749,14 @@ class Transfusion(Module):
         # add "sentence" start and end tokens when training
 
         if return_loss or need_velocity_matching:
-            for modality in modalities:
-                prepend(modality, tensor_([self.sos_id]))
-                modality.append(tensor_([self.eos_id]))
+            modalities = modalities.copy()
+
+            for i, modality in enumerate(modalities):
+                modalities[i] = [
+                    tensor_([self.sos_id]),
+                    *modality,
+                    tensor_([self.eos_id])
+                ]
 
         # need axial pos emb
 
@@ -1713,7 +1862,7 @@ class Transfusion(Module):
 
                 batch_modality_tokens.append(modality_tensor)
 
-                batch_modality_positions.append((modality_type, offset + 1, modality_length)) # offset + 1 due to extra [som] token
+                batch_modality_positions.append((modality_type, offset + prec_meta_tag_length, modality_length)) # offset + preceding meta tag length (which includes the modality start token)
 
                 offset += modality_length + 2 + len(modality_meta_info) + 1 # +2 due to [som] and [eom] - then account for meta start id and modality shape information (or eventually any meta information about modality)
 
@@ -1740,6 +1889,9 @@ class Transfusion(Module):
             modality_tokens.append(batch_modality_tokens)
             modality_positions.append(batch_modality_positions)
 
+        if return_loss:
+            total_tokens = sum([t.numel() for t in text])
+
         text = pad_sequence(text, padding_value = -1)
 
         if need_axial_pos_emb:
@@ -1784,8 +1936,6 @@ class Transfusion(Module):
 
         modality_positions, sorted_indices = order_modality_positions_by_seq_offset(modality_positions)
 
-        num_modalities = modality_positions.shape[-2]
-
         is_modalities = modality_positions_to_is_modality_mask(seq_len, modality_positions, num_modalities = self.num_modalities, device = device)
 
         is_any_modality = reduce(is_modalities, 'b t m n -> b n', 'any')
@@ -1910,10 +2060,6 @@ class Transfusion(Module):
         if return_only_pred_flows:
             return pred_flows
 
-        # calculate total tokens for weighing the loss
-
-        total_tokens = (text_labels != self.ignore_index).sum()
-
         # text autoregressive loss
 
         text_labels = text_labels.masked_fill(is_any_modality, self.ignore_index)
@@ -1929,6 +2075,8 @@ class Transfusion(Module):
         # calculate flow losses
 
         flow_losses = []
+        direction_losses = []
+
         modality_loss_weights = []
 
         for flow, pred_flow, is_one_modality in zip(flows, pred_flows, is_modalities.unbind(dim = 1)):
@@ -1943,12 +2091,19 @@ class Transfusion(Module):
 
             flow_loss = flow_loss[is_one_modality].mean()
 
+            if self.add_flow_direction_loss:
+                direction_loss = calc_direction_loss(pred_flow, flow)
+                direction_loss = direction_loss[is_one_modality].mean()
+
             modality_loss_weight = is_one_modality.sum() / total_tokens
 
             modality_loss_weights.append(modality_loss_weight)
 
             flow_losses.append(flow_loss)
 
+            if self.add_flow_direction_loss:
+                direction_losses.append(direction_loss)
+
         modality_loss_weights = torch.stack(modality_loss_weights)
 
         # only the token positions that are not modalities have autoregressive loss
@@ -1958,6 +2113,14 @@ class Transfusion(Module):
             (torch.stack(flow_losses) * modality_loss_weights).sum() * self.flow_loss_weight
         )
 
+        # maybe add direction loss
+
+        if self.add_flow_direction_loss:
+            total_loss = (
+                total_loss +
+                (torch.stack(direction_losses) * modality_loss_weights).sum() * self.direction_loss_weight
+            )
+
         # whether to handle velocity consistency
         # for straightening the flow, from consistency flow matching paper https://arxiv.org/abs/2407.02398
 
@@ -2000,4 +2163,4 @@ class Transfusion(Module):
         if not return_breakdown:
             return total_loss
 
-        return total_loss, LossBreakdown(total_loss, text_loss, flow_losses, velocity_match_losses)
+        return total_loss, LossBreakdown(total_loss, text_loss, flow_losses, velocity_match_losses, direction_losses)
